'' /////////////////////////////////////////////////////////////////////////////
'' Graphics Renderer for Donkey Kong Remake
'' (C) Steve Waddicor
'' //////////////////////////////////////////////////////////////////////////////

'TO-DO: Could have a second palette for preshifted pixel?

CON
  param_count   = 8
  NUMBER_OF_SPRITES       = 20

VAR
  long  cog
  long  renderer_ptr_block

PUB start(renderer_ptr) : okay

'' Start renderer - starts a cog
'' returns false if no cog available

  renderer_ptr_block := renderer_ptr
  okay := (cog := cognew(@entry,renderer_ptr)) > 0

DAT

'******************************
'* Assembly language renderer *
'******************************

                        org
'
'
' Entry
'
entry
                        rdlong  rend_cog_number,PAR
                        mov     current_line,rend_cog_number  'start rendering from the top

                        
do_frame                'Load parameters
                        mov     r0,PAR
                        movd    :load_dest,#renderer_params+1
                        mov     count,#param_count-1 + NUMBER_OF_SPRITES + NUMBER_OF_SPRITES
:load                   add     r0,#4
                        '       ___
:load_dest              rdlong  0-0,r0
                        '       ^^^ iterates through renderer_params array.
                        add     :load_dest,inc_dest
                        djnz    count,#:load

                        mov     x_tile,rend_side_scroll
                        and     x_tile,long_word_mask
                        mov     end_terrain_ptr,rend_terrain_ptr
                        add     end_terrain_ptr,terrain_width          
                        
do_line                 ' Draw a single scan line
                        mov     mix1,black
                        cmp     current_line,#31 wz,wc              
              if_be     jmp     #:after_colour_picked
                        cmp     current_line,#34 wz,wc
              if_ae     jmp     #:after_colour_picked
                        mov     mix1,colorbar71
:after_colour_picked
                        mov     mix2,mix1

                        mov     terrain_ptr,rend_terrain_ptr
                        add     terrain_ptr,x_tile
                        cmp     terrain_ptr,end_terrain_ptr wz,wc
              if_e      mov     terrain_ptr,rend_terrain_ptr

                        rdlong  terrain,terrain_ptr
                        add     terrain_ptr,#4
                        cmp     terrain_ptr,end_terrain_ptr wz,wc
              if_e      mov     terrain_ptr,rend_terrain_ptr

                        mov     x_pixel,rend_side_scroll
                        and     x_pixel,#%0000_0011 wc,wz

:init_loop    if_ne     rol     terrain,#8
              if_ne     djnz    x_pixel,#:init_loop

                        mov     x_pixel,rend_side_scroll
                        shl     x_pixel,#30

                        movd    scan_line_store,#cog_scanline_buffer

                        mov     p,#42
do_pixel
unroll_1
                        'get next pixel from terrain byte
                        mov     terrain_byte,terrain
                        and     terrain_byte,#$FF
                        cmp     current_line,terrain_byte wz

              if_e      or      mix2,#$6b 'Always or with black (or white when there is a smart bomb)
                        ror     mix2,#8
                        ror     terrain,#8
                        
                        add     x_pixel,x_pixel_inc wc
              if_nc     jmp     #:not_terrain_boundary
                        rdlong  terrain,terrain_ptr
                        add     terrain_ptr,#4
                        mov     x_pixel,#0
                        cmp     terrain_ptr,end_terrain_ptr wz,wc
              if_e      mov     terrain_ptr,rend_terrain_ptr
:not_terrain_boundary
                        'add     p,#1

unroll_2
                        'get next pixel from terrain byte
                        mov     terrain_byte,terrain
                        and     terrain_byte,#$FF
                        cmp     current_line,terrain_byte wz

              if_e      or      mix2,#$6b
                        ror     mix2,#8
                        ror     terrain,#8
                        
                        add     x_pixel,x_pixel_inc wc
              if_nc     jmp     #:not_terrain_boundary
                        rdlong  terrain,terrain_ptr
                        add     terrain_ptr,#4
                        mov     x_pixel,#0
                        cmp     terrain_ptr,end_terrain_ptr wz,wc
              if_e      mov     terrain_ptr,rend_terrain_ptr
:not_terrain_boundary
                        'add     p,#1

unroll_3
                        'get next pixel from terrain byte
                        mov     terrain_byte,terrain
                        and     terrain_byte,#$FF
                        cmp     current_line,terrain_byte wz

              if_e      or      mix2,#$6b
                        ror     mix2,#8
                        ror     terrain,#8
                        
                        add     x_pixel,x_pixel_inc wc
              if_nc     jmp     #:not_terrain_boundary
                        rdlong  terrain,terrain_ptr
                        add     terrain_ptr,#4
                        mov     x_pixel,#0
                        cmp     terrain_ptr,end_terrain_ptr wz,wc
              if_e      mov     terrain_ptr,rend_terrain_ptr
:not_terrain_boundary
                        'add     p,#1

unroll_4
                        'get next pixel from terrain byte
                        mov     terrain_byte,terrain
                        and     terrain_byte,#$FF
                        cmp     current_line,terrain_byte wz

              if_e      or      mix2,#$6b
                        ror     mix2,#8
                        ror     terrain,#8
                        
                        add     x_pixel,x_pixel_inc wc
              if_nc     jmp     #:not_terrain_boundary
                        rdlong  terrain,terrain_ptr
                        add     terrain_ptr,#4
                        cmp     terrain_ptr,end_terrain_ptr wz,wc
              if_e      mov     terrain_ptr,rend_terrain_ptr
:not_terrain_boundary
                        'add     p,#1
scan_line_store         mov     0-0,mix2    'dest starts at cog_scanline_buffer
                        add     scan_line_store,inc_dest 'self-modify dest
                        mov     mix2,mix1

                        djnz    p,#do_pixel

'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
                        ' Render sprites
                        mov     this_sprite,#rend_sprites_loc 'first sprites in array are drawn first
                        mov     this_sprite_tile,#rend_sprites_tile 'first sprites in array are drawn first
                        mov     outa,#0
sprite_loop
                        ' Verification: if the TV is already asking for this scanline or more
                        ' then panic, go to red alert with the debug LED and skip straight to TV output.
                        ' If the LED does light you need to consider optimising or allocating more
                        ' rendering cogs
                        call    #checktv
        if_ae           mov     outa,#1
        if_ae           jmp     #start_tv_copy

                        'get the sprite attibutes in sprite_attr which we'll use for the rest of the routine
                        movs    :sprite_attr_load,this_sprite
                        nop
:sprite_attr_load       mov     sprite_attr,0-0 wc
                        'don't draw it if it has the invisible flag
        if_c            jmp     #next_sprite

                        ' get the tile number and explode count
                        movs    :sprite_tile_load,this_sprite_tile
                        nop
:sprite_tile_load       mov     tile_number,0-0                                 ' EEEEEE.._........_........_TTTTTTTT
                        mov     explode_count,tile_number
                        shr     explode_count,#16
                        and     tile_number,#$FF

                        ' reject of x is out of range.
                        mov     sprite_x,sprite_attr          ' '%i.._YYYYY_YYYyyyyy_XXXXXXXX_XXXxxxxx (negative = invisible
                        shr     sprite_x,#5                   ' r0 = xpos
                        sub     sprite_x,#8                                     ' centre the sprite

                         ' find the y-span and reject if current-line isn't in it.
                        mov     r0,sprite_attr                                       'r0=sprite-y
                        shr     r0,#21
                        sub     r0,#4                                           ' centre the sprite

                        ' subtract explodecount*7 from x and explodecount*3 from y
                        add     r0,explode_count                                ' subtract explodecount*7 from x explodecount
                        add     sprite_x,explode_count                          
                        shl     explode_count,#3                                ' explodecount*8
                        sub     sprite_x,explode_count
                        shr     explode_count,#1
                        sub     r0,explode_count 
                        shr     explode_count,#1                                ' explodecount*4

                        and     r0,#$FF
                        
                        mov     draw_y,current_line
                        subs    draw_y,r0  wc,wz
              if_b      jmp     #next_sprite
                        mov     r1,#8                                           ' height of sprite in pixels
                        add     r1,explode_count                                ' explodecount*2
                        add     r1,explode_count                                ' explodecount*2
                        add     r1,explode_count                                ' explodecount*2
                        cmp     draw_y,r1 wc,wz
              if_ae     jmp     #next_sprite

                        and     sprite_x,world_width_mask
                        sub     sprite_x,rend_side_scroll
'                        cmp     sprite_x,screen_left wz,wc
'              if_ae     jmp     #:on_screen
'                        cmp     sprite_x,screen_right wz,wc
'              if_ae     jmp     #next_sprite
:on_screen
                        ' get the pixel data
                        shl     tile_number,#6          '64 bytes in a sprite
                        mov     hub_source,rend_sprite_table_ptr 'use hub_source for pixel data
                        add     hub_source,tile_number

                        ' add offset for y position in the tile
                        mov     r2,#0
                        mov     r3,#2
:adjust_draw_y_for_explode
                        cmp     draw_y,r2 wz,wc
              if_b      jmp     #next_sprite
                        cmp     draw_y,r3 wz,wc
              if_b      jmp     #:y_has_been_adjusted
                        sub     draw_y,explode_count wz,wc
              if_b      jmp     #next_sprite
                        add     r2,#2
                        add     r3,#2
                        jmp     #:adjust_draw_y_for_explode
                                      
:y_has_been_adjusted    shl     draw_y,#3                                       ' 8 bytes per sprite line
                        add     hub_source,draw_y

                        'calculate start address in scan line buffer        
                        mov     r0,sprite_x
                        shr     r0,#2                   'r0 = xpos/4
                        and     r0,#%1_1111_1111
                        mov     cog_dest,#cog_scanline_buffer
                        add     cog_dest,r0
                        
                        'calculate where the sprite stars within a frame
                        and     sprite_x,#%0000_0011
                        'mov     sprite_x,#3

'***************************************************************************************************
                        'Code for converting 16 color sprites to hi-color

                        mov     count3,#2
                         ' put pixel colours in own longs
:source_long_loop       rdlong  r1,hub_source           'pixels(r1) = $01234567
                        add     hub_source,#4           'after first long, do second long.

                        mov     count2,#4
:source_byte_loop
                        'do a bounds check on scan line line buffer
                        cmp     cog_dest,#cog_scanline_buffer wz,wc
              if_b      jmp     #:common2
                        cmp     cog_dest,#end_of_cog_scanline_buffer wz,wc
              if_ae     jmp     #:common2
                        
                        rol     r1,#8
                        mov     r2,r1
                        and     r2,#$FF wz              'first 2 pixels
              if_z      jmp     #:common2
              
                        mov     r3,r2                   

                        shr     r2,#4                   'r2 first pixel to display
                        add     r2,#palette             'r2 is palette pointer
                        movs    :lookup_color_1,r2
                        
                        and     r3,#$F                  'r3 is second pixel to display
                        add     r3,#palette
                        movs    :lookup_color_2,r3

:lookup_color_1         mov     r2,0-0
:lookup_color_2         mov     r3,0-0

                        'mov     r2,blah3
                        'mov     r3,blah3


                        cmp     sprite_x,#1 wz,wc
              if_e      jmp     #:offset1
                        cmp     sprite_x,#2 wz,wc
              if_e      jmp     #:offset2
                        cmp     sprite_x,#3 wz,wc
              if_e      jmp     #:offset3

:offset0
                        'shl     r2,#24
                        'shl     r3,#16
                        shl     r3,#8

                        movs    :maskit,#offset0mask
                        movd    :maskit,cog_dest
                        movd    :drawit,cog_dest
                        jmp     #:common

:offset1
                        shl     r3,#16
                        shl     r2,#8
                        movs    :maskit,#offset1mask
                        movd    :maskit,cog_dest
                        movd    :drawit,cog_dest
                        jmp     #:common
:offset2
'                        shl     r2,#8
                        shl     r3,#24
                        shl     r2,#16 
                        movs    :maskit,#offset2mask
                        movd    :maskit,cog_dest
                        movd    :drawit,cog_dest
:common
                        or      r2,r3
:maskit                 and     0-0,0-0
:drawit                 or      0-0,r2
                        jmp     #:common2

:offset3
                        shl     r2,#24
                        movd    :maskit1,cog_dest
                        movd    :drawit1,cog_dest
                        add     cog_dest,#1
                        movd    :maskit2,cog_dest
                        movd    :drawit2,cog_dest
                        sub     cog_dest,#1
                                          
:maskit1                and     0-0,offset3mask1
:maskit2                and     0-0,offset3mask2
:drawit1                or      0-0,r2
:drawit2                or      0-0,r3

:common2                'shr     r1,#8 
                        add     sprite_x,#2
                        add     sprite_x,explode_count
:adjust_x               cmp     sprite_x,#4 wz,wc
              if_ae     sub     sprite_x,#4
              if_ae     add     cog_dest,#1
              if_ae     jmp     #:adjust_x
              
                        djnz    count2,#:source_byte_loop
                        djnz    count3,#:source_long_loop


next_sprite             add     this_sprite,#1
                        add     this_sprite_tile,#1
                        cmp     this_sprite,#rend_sprites_end wz
              if_ne     jmp     #sprite_loop

'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

                        ' We've finished rendering the line and we didn't have a hissy fit panic.
                        ' We need to wait here for the line we just prepared to be requested by the TV driver
wait_for_the_request
                        call    #checktv
              if_b      jmp     #wait_for_the_request

                        ' Our line has been requested so copy it to hub scan line buffer
start_tv_copy           movs    :nextcopy, #cog_scanline_buffer
                        mov     r0, rend_scanline_buffer
                        mov     count, #73              ' length of visible scanline buffer in longs
:nextcopy               mov     r1,cog_scanline_buffer
                        add     :nextcopy, #1
                        wrlong  r1, r0
                        add     r0, #4
                        djnz    count, #:nextcopy

                        ' We've handed over the scanline to the TV driver, so do another.
allocate_next_job       add     current_line,rend_cog_count 'we only do every Nth line
                        ' Degrade gracefully when there's too much work to do.
'                        call    #checktv
'              if_ae     jmp     #allocate_next_job
                       ' Wrap around.
                        cmp     current_line,#224 wz,wc  'cmpsub will subtract 224 if current_line is greater than 224.
              if_b      jmp     #do_line
                        sub     current_line,#224
                        jmp     #do_frame


'-------------------------------------------------------------------
checktv                 ' Tf we're currently rendering a low numbered line when a high one is being requested
                        ' then that's OK because we wrapped around.



                        rdlong  scanline_req,rend_scanline_req_adr
                        cmp     scanline_req,#210 wc,wz
              if_be     jmp     #:checkit
                        cmp     current_line,#7 wc,wz
              if_be     cmp     scanline_req, current_line wz
                        'b = normal
                        'z = doit
                        'a = err
              if_be     jmp     #checktv_ret

:checkit                cmp     scanline_req, current_line wz, wc    'set carry if request is less than current.
                        'b = normal
                        'z = doit
                        'a = err

                        ' At this point, we could skip drawing sprites or something like that and go
                        ' straight to image output and still stay synced with the tv driver
                        ' We must be quick though!
                        'jmp #start_tv_copy
checktv_ret             ret
'-------------------------------------------------------------------

' Initialized data
inc_dest                long    1 << 9 << 0
inc_dest_2              long    1 << 9 << 1
attr_invisible          long    $0000_8000
constantFF00            long    $FF00
black                   long    $02020202
colorbar71              long    $0b_0b_0b_0b

DAT 'Uninitiaised data - always keep it at the end of file to avoid nasty gotcha with RES.

'sprites
palette                 long    $00000002 'black
                        long    $0000001c 'purple (ship wing)
                        long    $0000002d
                        long    $00000006 'white
                        long    $000000Bc 'green
                        long    $0000007d 'yellow
                        long    $0000006d
                        long    $0000007d
                        long    $0000008d
                        long    $0000009d
                        long    $000000ad
                        long    $000000bd
                        long    $000000cd
                        long    $000000dd
                        long    $000000ed
                        long    $000000fd

' New for Defender
world_width_mask        long    %111_1111_1111
screen_left             long    -64
screen_right            long    292
'top_three_bytes         long    $FFFF_FF00
long_word_mask          long    $FFFF_FFFC
terrain_width           long    2048
x_pixel_inc             long    1<<30

'offset0mask             long    $0000FFFF
'offset1mask             long    $FF0000FF
'offset2mask             long    $FFFF0000
'offset3mask1            long    $FFFFFF00
'offset3mask2            long    $00FFFFFF

offset0mask             long    $FFFF0000
offset1mask             long    $FF0000FF
offset2mask             long    $0000FFFF
offset3mask2            long    $FFFFFF00
offset3mask1            long    $00FFFFFF

blah1                    long    $ab
blah2                    long    $6b
blah3                   long    $6b6b6b6b

' Parameter buffer
renderer_params
rend_cog_number         res     1
rend_scanline_req_adr   res     1
rend_sprite_table_ptr   res     1
rend_color_table_ptr    res     1               'rend_colors
rend_scanline_buffer    res     1
rend_cog_count          res     1
rend_terrain_ptr        res     1
rend_side_scroll        res     1

rend_sprites_loc        res     NUMBER_OF_SPRITES
rend_sprites_end                
rend_sprites_tile       res     NUMBER_OF_SPRITES
' End of parameter buffer

' Cog scanline buffer = 292 pixels, 8 bits per pixel = 2336 bits = 292 bytes = 73 longs
scratch1                res     16
cog_scanline_buffer     res     42
end_of_cog_scanline_buffer
scratch2                res     16
' End of scanline buffer


count                   res     1
count2                  res     1
count3                  res     1
hub_map_ptr             res     1
current_line            res     1
tile_table              res     1

draw_y                  res     1

sprite_x                res     1
sprite_attr             res     1
this_sprite             res     1
scanline_req            res     1
x                       res     1
tile                    res     1
pixels1                 res     1
pixels2                 res     1
mix1            
shifted1                res     1
mix2          
shifted2                res     1
shifted3                res     1
shifted4                res     1
shifted5                res     1
r0                      res     1
r1                      res     1
r2                      res     1
r3                      res     1
tile_number             res     1
hub_source              res     1
cog_dest                res     1

pixel_color_list1       res     1
pixel_color_list2       res     1
pixel_color_list3       res     1
pixel_color_list4       res     1

pixel_mask_list1        res     1
pixel_mask_list2        res     1
pixel_mask_list3        res     1
pixel_mask_list4        res     1

shifted_mask1           res     1
shifted_mask2           res     1
shifted_mask3           res     1
shifted_mask4           res     1
shifted_mask5           res     1

last_tile_palette       res     1
last_tile               res     1

'new for Defender
this_sprite_tile        res     1
terrain_ptr             res     1
end_terrain_ptr         res     1
terrain                 res     1
terrain_byte            res     1
p                       res     1
x_tile                  res     1
x_pixel                 res     1
explode_count           res     1       
